#
# koji_utils.py - helper routines for dealing with koji builds
#
# Copyright 2009, Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Author: Will Woods <wwoods@redhat.com>

import time
import koji
from repoinfo import repoinfo
import rpmUtils.miscutils
import sys
import autoqa.config
from autoqa.config import autoqa_conf
import os
import autoqa.util as util
import xmlrpclib

kojiurl = autoqa_conf.get('resources', 'koji_url')
pkgurl = autoqa_conf.get('resources', 'pkg_url')

class SimpleKojiClientSession(koji.ClientSession):
    '''Convenience wrapper class for interacting with koji'''

    def __init__(self, server=kojiurl, opts=None):
        return koji.ClientSession.__init__(self, server, opts)


    def list_builds_since(self, timestamp):
        '''Return a list of new builds since the given timestamp'''
        return self.listBuilds(completeAfter=timestamp)


    def list_tags(self, nvr):
        '''Return a list of tags applied to the package with the given nvr'''
        return [t['name'] for t in self.listTags(nvr)]


    def tag_history(self, nvr):
        '''Returns a list of every tag ever applied to the given nvr'''
        return [t['tag_name'] for t in self.tagHistory(build=nvr)]


    def latest_by_tag(self, tag, pkgname, max_evr=None):
        '''Get the latest package for the given name in the given tag. If you
        set max_evr, it is *exclusive*.
        max_evr = (epoch, version, release)
        '''
        # allow epoch to be empty, transcode it to zero
        if max_evr and not max_evr[0]:
            max_evr = (0, max_evr[1], max_evr[2])

        if max_evr:
            pkgs = self.listTagged(tag, package=pkgname)
            # sort the packages from highest EVR to lowest
            pkgs.sort(cmp=lambda x, y: compareEVR(x, y), reverse=True)
            for pkg in pkgs[:]:
                if rpmUtils.miscutils.compareEVR((pkg['epoch'] or '0',
                pkg['version'], pkg['release']), max_evr) >= 0:
                    # remove all packages greater or equal than max_evr
                    pkgs.remove(pkg)
                else:
                    # since packages are sorted, we can stop here
                    break
        else:
            pkgs = self.listTagged(tag, package=pkgname, latest=True)
        if pkgs:
            # the winner is the first package in list
            return pkgs[0]


    def build_failed(self, nvr):
        return (self.getBuild(nvr)['state'] > 1)


    # Idea taken from the koji CLI
    def check_connection(self):
        '''Check the connection to koji.'''
        try:
            ret = self.getAPIVersion()
        except xmlrpclib.ProtocolError:
            print >> sys.stderr, "Unable to connect to Koji server"
            raise
        if ret != koji.API_VERSION:
            print >> sys.stderr, "Warning: Koji API mismatch (Server %d, client %d)" % (ret,koji.API_VERSION)


    def test_list_builds_since(self):
        builds = []
        starttime = time.time()
        timestamp = time.time()
        while not builds and (starttime - timestamp < 10000):
            timestamp -= 3000
            builds = self.list_builds_since(timestamp)
        print "Most recent builds: %.0f min ago" % ((starttime-timestamp)/60.0)
        return builds


    def list_previous_releases(self, name, tag, max_evr=None,
        unstable_tags=False):
        '''Get most recent releases of package in specified tag (inheritence
        enabled). Returns dictionary in the format {tag: NVR object}. Optional
        max_evr says that all returned release versions must be lesser then
        it. If you enable unstable_tags, then also '-testing' tags are considered.

        max_evr = (epoch, version, release)
        '''
        prev_rlss = {}
        # minor hack - assume -candidate parents = stable+updates+testing
        # TODO: provide this info in repoinfo instead
        # since there's no repo for candidate builds, just look at -testing and below
        if tag.endswith('-updates-candidate'):
            tag = tag.replace('-candidate','-testing')

        testrepo = repoinfo.getrepo_by_tag(tag)
        if testrepo is None:
            print >> sys.stderr, "Error: There's no repo definition for tag '%s'" % tag
            return {}
        parents = testrepo['parents']
        repos = parents + [testrepo['name']]

        if not unstable_tags:
            for repo in repos[:]:
                rtag = repoinfo.get(repo, 'tag')
                if ('-testing' in rtag or '-candidate' in rtag):
                    # remove repo with unstable tag
                    repos.remove(repo)

        for repo in repos:
            rtag = repoinfo.get(repo, 'tag')
            prev_rlss[rtag] = self.latest_by_tag(rtag, name, max_evr=max_evr)
        return prev_rlss


    def list_previous_release(self, name, tag, max_evr=None,
        unstable_tags=False):
        ''' The same as list_previous_releases(), but returns a single NVR
        object (the most recent one) regardless of its tag.
        '''
        prev_rlss = self.list_previous_releases(name, tag, max_evr,
            unstable_tags)
        # take only NVR objects and erase possible None's
        prev_rlss = [rls for rls in prev_rlss.values() if rls]
        prev_rlss.sort(cmp=lambda x, y: compareEVR(x, y), reverse=True)
        if prev_rlss:
            return prev_rlss[0]


    def nvr_to_urls(self, nvr, arches=None, debuginfo=False, src=True):
        '''Get list of URLs for RPMs corresponding to a build.
        @param nvr build NVR
        @param arches restrict the arches of builds to provide URLs for. If
                      basearch i386 is in the list, i686 arch is automatically
                      added (since that's the arch Koji API uses).
        '''

        # add i686 arch if i386 is present in arches
        if arches and 'i386' in arches and 'i686' not in arches:
            arches.append('i686')

        info = self.getBuild(nvr)
        baseurl = '/'.join((pkgurl, info['package_name'],
                           info['version'], info['release']))

        rpms = self.listRPMs(buildID=info['id'], arches=arches)
        if not debuginfo:
            rpms = filter(lambda r: not r['name'].endswith('-debuginfo'), rpms)
        if not src:
            rpms = filter(lambda r: not r['arch'] == 'src', rpms)

        urls = ['%s/%s' % (baseurl, koji.pathinfo.rpm(r)) for r in rpms]
        return sorted(urls)


    def nvr_to_rpms(self, nvr, debuginfo=False, src=True):
        ''' Get a list of RPMs existing for specified NVR. All RPMs have 'url'
        key added with URL where to download the RPM as the value.'''

        info = self.getBuild(nvr)
        baseurl = '/'.join((pkgurl, info['package_name'],
                           info['version'], info['release']))

        rpms = self.listRPMs(buildID=info['id'])
        if not debuginfo:
            rpms = filter(lambda r: not r['name'].endswith('-debuginfo'), rpms)
        if not src:
            rpms = filter(lambda r: not r['arch'] == 'src', rpms)

        for r in rpms:
            r['url'] = '%s/%s' % (baseurl, koji.pathinfo.rpm(r))

        return rpms


    def get_nvr_rpms(self, nvr, rpm_dir, arches=None, debuginfo=False, src=False,
                     allow_cache=True):
        '''Retrieve the RPMs associated with a build NVR into the specified
           directory.

        @param nvr build NVR
        @param rpm_dir location where to store the RPMs
        @param allow_cache if true and koji_pkg_cache in autoqa.conf is true,
            then rpm_dir value is ignored and all the RPMs are downloaded
            to a persistent cache storage and survive test end; otherwise
            rpm_dir value is honored
        @return list of local filenames of the grabbed RPMs
        @raise urlgrabber.grabber.URLGrabError if downloading failed
        '''

        rpm_urls = self.nvr_to_urls(nvr, arches=arches, debuginfo=debuginfo, src=src)
        rpm_files = []
        print 'Fetching RPMs for: %s' % nvr
        for url in rpm_urls:
            rpm_file = self.download_rpm(url, rpm_dir, allow_cache)
            rpm_files.append(rpm_file)

        return rpm_files


    def download_rpm(self, url, rpm_dir, allow_cache=True):
        '''Download a single RPM.

        @param url url of the RPM
        @param rpm_dir location where to store the RPM
        @param allow_cache if true and koji_pkg_cache in autoqa.conf is true,
            then @param rpm_dir is ignored and the RPM is downloaded to a
            persistent cache storage and survive test end; otherwise @param
            rpm_dir is honored
        @return local filename of the grabbed RPM
        @raise urlgrabber.grabber.URLGrabError if downloading failed
        '''

        # if we want to cache packages (both in config and here), change the
        # rpm_dir destination
        use_cache = allow_cache and autoqa.config.getbool(autoqa_conf.get(
                                    'resources', 'koji_pkg_cache'))
        if use_cache:
            rpm_dir = autoqa.config.rpmdir

        # create destination dir if needed
        if not os.path.exists(rpm_dir):
            os.makedirs(rpm_dir)

        assert os.path.isdir(rpm_dir)

        # download
        overwrite = not allow_cache
        rpm_file = util.download(url, rpm_dir, overwrite)

        return rpm_file


def compareEVR(nvr1, nvr2):
    '''Compare EVR from two NVR objects.
    @param nvr1 Koji build object or None
    @param nvr2 Koji build object or None
    @return The higher of the two builds when compared according to EVR. If one
            of the objects is None, then it's always lesser. If both are None,
            then they are equal.
    '''
    # handle empty input
    if not nvr1:
        return 0 if not nvr2 else -1
    if not nvr2:
        return 1
    # compare
    return rpmUtils.miscutils.compareEVR((nvr1['epoch'] or '0',
        nvr1['version'], nvr1['release']), (nvr2['epoch'] or '0',
        nvr2['version'], nvr2['release']))


def getENVR(build):
    ''' Get NVR (if epoch is 0) or ENVR (otherwise) from build info
    @param build buildinfo dictionary (as returned from koji methods like getBuild)
    @return string NVR or ENVR '''
    if build['epoch']:
        return '%s:%s' % (build['epoch'], build['nvr'])
    else:
        return build['nvr']
