#
# Copyright 2009, Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Author: James Laska <jlaska@redhat.com>

import os
import os.path
import re
import grp, stat
import StringIO
import urlgrabber
import ConfigParser
from autoqa.repoinfo import repoinfo
from autoqa.test import AutoQATest
from autoqa.decorators import ExceptionCatcher
from autotest_lib.client.bin import utils
from autotest_lib.client.common_lib import error

class anaconda_checkbot(AutoQATest):
    version = 1 # increment if setup() changes


    @ExceptionCatcher()
    def setup(self, **kwargs):
        super(self.__class__, self).setup()
        utils.system('yum -y install mock')
        # If autotest isn't already a member of the mock group, add it now
        utils.system('usermod -a -G mock autotest')


    @ExceptionCatcher()
    def initialize(self, config, **kwargs):
        super(self.__class__, self).initialize(config)

    @ExceptionCatcher()
    def run_once(self, baseurl, event, arch, *args, **kwargs):
        super(self.__class__, self).run_once()

        assert event in ['post-tree-compose', 'git-post-receive'], \
            'Unexpected event argument: %s' % event

        # XXX: The methods used to detect the releasever, basearch and
        # gitbranch are a complete hack.  The method takes the repo URL
        # provided and attempts to guess apropriate values by first inspecting
        # .treeinfo.  If no .treeinfo is found, we resort to pattern matching on
        # the URL.  If that fails, we default to: releasever='fedora-rawhide',
        # basearch=$(uname -i), gitbranch='master'.  Long-term, this will be
        # simpler when a 'post-git-update' watcher is available.  Until then,
        # enjoy!

        if event == 'git-post-receive':
            gitbranch = kwargs.get('branch','master')
            basearch = arch

            # Guess the releasever using the branch as a guide "f<digit>" or "rhel<digit>"
            m = re.search(r'^\b(rhel|epel|f)([0-9]+)\b', gitbranch, re.IGNORECASE)
            if m and m.group(1).startswith('f'):
                releasever = 'fedora-%s' % m.group(2)
            elif m and m.group(1) in ['epel','rhel']:
                releasever = 'epel-%s' % m.group(2)
            else:
                releasever = 'fedora-rawhide'
        else:
            # Download .treeinfo from the provided baseurl
            data = StringIO.StringIO()
            try:
                grabber = urlgrabber.grabber.URLGrabber(prefix=baseurl)
                data.write(grabber.urlread('.treeinfo'))
            except urlgrabber.grabber.URLGrabError, e:
                self.log('Failed to download .treeinfo: %s' % str(e),
                         stderr=True, highlight=True)

            # Create ConfigParser object using .treeinfo data
            c = ConfigParser.ConfigParser()
            treeinfo = c.readfp(data)

            # Determine appropriate releasever and branch (default to rawhide)
            if c.has_option('general', 'version') and \
               c.has_option('general', 'family'):
                if c.get('general', 'family').startswith('Fedora'):
                    releasever = 'fedora-%d' % float(c.get('general', 'version'))
                    gitbranch = 'f%d-branch' % float(c.get('general', 'version'))
                elif c.get('general', 'family').startswith('Red Hat'):
                    releasever = 'epel-%d' % float(c.get('general', 'version'))
                    gitbranch = 'rhel%d-branch' % float(c.get('general', 'version'))
            else:
                # Ugly, but guess from the URL ...
                m_fedora = re.search(r'\b(fedora)\b.*\b([0-9]+)\b',
                    baseurl, re.IGNORECASE)
                m_epel = re.search(r'\b(rhel|epel)-([0-9])\b',
                    baseurl, re.IGNORECASE)
                if m_fedora:
                    releasever = 'fedora-%s' % m_fedora.group(2)
                    gitbranch = 'f%s-branch' % m_fedora.group(2)
                elif m_epel:
                    releasever = 'epel-%s' % m_epel.group(2)
                    gitbranch = 'rhel%s-branch' % m_epel.group(2)
                else:
                    releasever = 'fedora-rawhide'
                    gitbranch = 'master'
                    self.log('Unable to detect test release, using system ' +
                             'release: %s' % releasever, stderr=True,
                             highlight=True)

            # Determine appropriate basearch (default to uname -i)
            if c.has_option('general', 'arch'):
                basearch = c.get('general', 'arch')
            else:
                # Ugly, but guess from the URL ...
                m = re.search(r'\b(i[0-9]86|x86_64|ppc[0-9]*|s390x?)\b', \
                              baseurl)
                if m:
                    basearch = m.group(1)
                else:
                    basearch = arch
                    self.log('Unable to detect test architecture, using ' +
                             'system architecture: %s' % basearch, stderr=True,
                             highlight=True)

        # Adjust permissions so autotest can save results to self.resultsdir
        os.chown(self.resultsdir, -1, grp.getgrnam('autotest').gr_gid) # set group = 'autotest'
        os.chmod(self.resultsdir, os.stat(self.resultsdir).st_mode | stat.S_IWGRP) # allow group write

        # Allow autotest user to create a mock config
        mockdir_stat = os.stat('/etc/mock') # st_mode, st_uid, st_gid
        os.chown('/etc/mock', -1, grp.getgrnam('autotest').gr_gid) # set group = 'autotest'
        os.chmod('/etc/mock', mockdir_stat.st_mode | stat.S_IWGRP) # allow group write

        # Run checkbot
        cmd = '%s/checkbot.sh -r %s -a %s -d %s -b %s -m %s' % (self.bindir, \
              releasever, basearch,self.resultsdir, gitbranch, \
              os.path.join(self.bindir, 'mock.cfg.in'))
        cmd = 'su -c "%s" - autotest' % cmd
        try:
            out = utils.system_output(cmd, retain_output=True)
        except error.CmdError, e:
            self.detail.result = 'FAILED'
            out = e.result_obj.stdout
        finally:
            self.log(out, printout=False)
            # Reset permissions on '/etc/mock' directory
            os.chown('/etc/mock', -1, mockdir_stat.st_gid) # reset group id
            os.chmod('/etc/mock', mockdir_stat.st_mode) # reset mode

        # Attempt to summarize test results
        self.detail.id = gitbranch
        summary_detail = ''
        for log in ['pylint-log', 'unittest.log', 'build.log', 'rpmbuild.log']:
            logfile = os.path.join(self.resultsdir, log)
            if not os.path.isfile(logfile):
                summary_detail += ' * %s - FAILED (%s not found)\n' % \
                    (os.path.splitext(log)[0], log)
                self.detail.result = 'FAILED'
            else:
                # Read test log
                fd = open(logfile, 'r')
                data = fd.read()
                fd.close()

                # Summarize pylint result
                if log == 'pylint-log':
                    # pylint displays errors in the format: <type><code>, where
                    # <type> is a single character, and <code> represents a
                    # specific error code.  The following scans pylint output
                    # and sums up the different error <type>'s
                    err_cnt = dict(C=0, R=0, W=0, E=0, F=0)
                    for line in data.split('\n'):
                        m = re.search(r'^([CRWEF])[0-9]+:', line)
                        if m:
                            ecode = m.group(1)
                            if not err_cnt.has_key(ecode):
                                err_cnt[ecode] = 0
                            err_cnt[ecode] += 1
                    summary_detail += ' * %s - %s\n' % (os.path.splitext(log)[0], \
                                       ' '.join(['%s%s' % p for p in err_cnt.items()]))
                # Summarize unittest result
                elif log == 'unittest.log':
                    (okay, errors) = (0, 0)
                    # Count lines ending in 'ok' or 'ERROR' as test results
                    for line in data.split('\n'):
                        if re.search(r'ok$', line): okay += 1
                        if re.search(r'ERROR$', line): errors += 1
                    summary_detail += ' * %s - %s - %s of %s tests passed\n' % \
                        (os.path.splitext(log)[0], \
                         (errors==0 and okay>0) and "OK" or "FAILED",\
                         okay, errors+okay)

                # Summarize anaconda-build result
                elif log == 'build.log':
                    # make provides status by way of the exit code, but we
                    # can't capture that since 'checkbot.sh' is calling make.
                    # Instead, count any occurances of '***' as a failure.  For
                    # example ...
                    #  make[1]: Entering directory `/home/clumens/src/anaconda'
                    #  make[1]: *** No rule to make target `am--refresh'.  Stop.
                    #  make[1]: Leaving directory `/home/clumens/src/anaconda'
                    #  make: *** [../aclocal.m4] Error 2
                    num_errors = data.count('***')
                    summary_detail += ' * %s - %s\n' % (os.path.splitext(log)[0], \
                                    num_errors==0 and "OK" \
                                    or "FAILED (%s errors)" % num_errors)
                # Summarize rpmbuild.log result
                elif log == 'rpmbuild.log':
                    # Scan the rpmbuild.log to determine success or failure
                    m = re.search(r'^error:\s*(.*)$', data, re.MULTILINE)
                    if m:
                        summary_detail += ' * %s - FAILED - %s\n' % \
                            (os.path.splitext(log)[0], m.group(1))
                    else:
                        m = re.findall(r'^Wrote:\s*(.*\.rpm).*$', data, re.MULTILINE)
                        # If packages were built, that's usually a good thing
                        if len(m) > 0:
                            summary_detail += ' * %s - OK' % os.path.splitext(log)[0]
                        else:
                            summary_detail += ' * %s - FAILED' % os.path.splitext(log)[0]
                        summary_detail += ' - built %s rpms\n' % len(m)
                else:
                    summary_detail += ' * %s - NEEDS_INSPECTION\n' % (os.path.splitext(log)[0],)

                # Append complete log file to output
                self.log('== %s ==\n%s\n' % (log, data))

        self.detail.log.summary_detail = summary_detail


        distnum = releasever.split('-')[1]
        # now in repoinfo.conf is no repo for epel
        distro = 'f' if releasever.startswith('f') else 'el'

        reponame = 'rawhide' if (distnum == 'rawhide') else distro+distnum

        if (reponame in repoinfo.repos()):
            kojitag = repoinfo.get(reponame, 'tag')
            self.post_results(self.detail, optin={'pkg': 'anaconda', 'kojitag': kojitag})
        # if anaconda team will want to sent email AND opt-in emails we have to
        # add
        # else:
        #     kojitag = 'uknown'
        # self.post_results(optin={'pkg': 'anaconda', 'kojitag': kojitag})
        # and we have to erase a line above this commentary
