#!/usr/bin/python
#
# Copyright 2010, Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Chris Lumens <clumens@redhat.com>
"""
Base classes for creating test suites.  A storage test suite consists of a
single class, which is a subclass of BaseSuite.

A test suite is an abstraction on top of test cases.  Each suite contains at
least one, but potentially many, test cases and runs them in sequence.  In
this way, much more complicated test setups can be created by putting together
the pieces.  For instance, one test case can initialize one disk image, a
second test case can initialize a second disk image, and a third test case can
test what happens when attempting to use both in some new way.

The results for all test cases in the suite will be reported in case something
goes wrong in an intermediate step.  However, only the results for the last
case should be considered as real proof of if the entire suite passed or not.
"""

import glob
import itertools
import logging
import os
import pickle
import shutil
import string
import tempfile
from threading import Timer

class BaseSuite(object):
    """A BaseSuite instance implements a test suite.  Creating a subclass of
       this top-level object is pretty simple, though.  All suites must provide
       a tests property, and that is it.  All other methods should not be
       overridden.

       Class attributes:

       name -- A name for this test suite.  All test suites must have a name.
               This string should not contain spaces or other unusual
               characters as it will be used for creating directories to
               hold results and temporary images.
       timeout -- How much time (in seconds) should be allotted for this test
                  suite to run.  When the timeout is reached, the suite will
                  be killed and marked as failed.  Subclasses may set whatever
                  timer they feel is appropriate.
    """

    name = "base"
    timeout = 5*60

    def __init__(self, basedir, diskimagesdir):
        """Create a new BaseSuite instance.

           drives -- A list of filenames to disk images that will be used
                     in all tests in this suite.  The setup() method must
                     create these images and populate this list.  It is not
                     necessary for test cases to use every single drive in
                     this list.
        """
        self.basedir = basedir
        self.diskimagesdir = diskimagesdir
        self.drives = []

        self._runningReportDir = None
        self._runningTest = None
        self._tempdir = None

    def cleanup(self):
        """Destroy everything created by all test cases.  This is done after
           all cases have been run, since it might involve removing disk images
           that will be needed by later cases.  Test suites should not override
           this method without also later calling this base method.
        """
        shutil.rmtree(self.tempdir, ignore_errors=True)

    def die(self, *args, **kwargs):
        """Kill whatever test case is currently running and mark it as failed."""
        logging.error("Timer expired, killing test suite %s" % self.name)
        self._runningTest.die()

        # Mark the running test as failed.
        if not os.path.exists(self._runningReportDir):
            os.makedirs(self._runningReportDir)

        # If the test was previously marked as passed but since timed out,
        # then it really failed.
        if os.path.exists(self._runningReportDir + "/PASSED"):
            os.unlink(self._runningReportDir + "/PASSED")

        open(self._runningReportDir + "/FAILED", "w").close()

    @property
    def runnable(self):
        """If this method returns False, the suite will not be run by the test
           framework.  Test suites can override this method as needed.
        """
        return True

    def run(self, token, livecd, resultsdir):
        """Execute this test suite.  This is done by iterating over all test
           case objects, writing out config files, and then running their
           own run() methods.  Test suites should never override this method.
        """
        driverIn = self.basedir + "/driver.py.in"
        driverOut = driverIn.replace(".in", "")

        # Start a timer now.  If all test cases do not complete in time,
        # terminate the suite.  Otherwise we cancel the timer later.
        timer = Timer(self.timeout, self.die)
        timer.start()

        for (obj, ndx) in zip(self.tests, itertools.count(1)):
            # We need a reference to the running test case in case the timer expires.
            self._runningTest = obj

            # Read in the driver.py template, do a textual substitution, then
            # write it back out.
            fo = open(driverIn, "r")
            s = "".join(fo.readlines())
            fo.close()

            fo = open(driverOut, "w")
            fmt = string.Formatter()

            reportSubdir = self.name + "/" + str(ndx) + "-" + obj.name
            self._runningReportDir = resultsdir + "/" + reportSubdir

            fo.write(fmt.format(s, module=obj.name.replace("TestSuite", ""),
                                   reportDir=reportSubdir,
                                   runnerClass=obj.name + "Runner",
                                   runnerClassAttrs=pickle.dumps(obj.runnerClassAttrs),
                                   testClass=obj.name + "TestCase",
                                   testClassAttrs=pickle.dumps(obj.testClassAttrs),
                                   token=token))
            fo.close()
            os.chmod(driverOut, 0755)

            logging.info("----- Creating Test Case %s-%s -----" % (str(ndx), obj.name))
            obj.create()
            logging.info("----- Running Test Case %s-%s -----" % (str(ndx), obj.name))
            obj.run(livecd)

            # If the test failed, there's no point in continuing on to the
            # next case in this suite.  We might as well terminate now and save
            # some time.
            if os.path.exists(self._runningReportDir + "/FAILED"):
                timer.cancel()
                return

        timer.cancel()

    def setup(self):
        """Create all disk images to be used in this test suite, populating the
           self.drives list along the way.
        """
        if self.__class__ is BaseSuite:
            raise NotImplementedError, "setup() not implemented for BaseSuite"

    @property
    def tempdir(self):
        """Return the directory that will be used to store all temporary data
           for this test suite.  Creating this directory is done at the suite
           level, not at the test case level.
        """
        if not self._tempdir:
            self._tempdir = tempfile.mkdtemp(prefix="%s-" % self.name, dir=self.diskimagesdir)

        return self._tempdir

    @property
    def tests(self):
        """Return a list of cases.Creator instances, each corresponding to one
           test case in this suite.  These will be executed in the order
           provided and each will only make use of the disk images given as an
           argument to each Creator instance.
        """
        return []
